% gausssrc.m
% tests the SAM algorithm and vector CMA for equalization
% of a Gaussian source (in a single carrier system)
%
% Written by Rick Martin, 06/06/2002
% Last updated 06/19/2002
% Cornell University Blind Equalization Research Group
%
% Copyright 2002 Cornell University, All Rights Reserved
%
% SOFTWARE shall mean "gausssrc.m", Cornell Research Foundation 
% ("CRF") Docket D-3006.  SOFTWARE is made available to allow
% non-commercial and research use of SOFTWARE.  Cornell
% University specifically reserves all commercial rights to
% SOFTWARE and these rights may be licensed by CRF on behalf   
% of Cornell University to third parties.
%
% CRF provides SOFTWARE on an ``as is'' basis.  CRF does not
% warrant, guarantee, or make any representations regarding 
% the use or results of SOFTWARE with respect to its
% correctness, accuracy, reliability or performance.  The
% entire risk of the use and performance of SOFTWARE is  
% assumed by EVALUATOR.  ALL WARRANTIES INCLUDING, WITHOUT
% LIMITATION, ANY WARRANTY OF FITNESS  FOR A PARTICULAR   
% PURPOSE OR MERCHANTABILITY ARE HEREBY EXCLUDED.

% system parameters
c = [0.15 1 0.2 -0.1 0.1];  % channel
%c = [1 1/2 1/4 1/8 1/16];
%c = [1 0.5];
c = c/norm(c);
Lc = length(c)-1;     % channel length - 1
sx = 1;               % sqrt(source power)
ssx = sx*sx;          % source power
SNR = 40;             % in dB
sn = sx*10^(-SNR/20); % sqrt(noise power)
ssn = sn*sn;          % noise power
rho = 3;              % E(source^4)/E(source^2), for VCMA

% algorithm parameters
iter = 100000;        % number of iterations
cnst = 0;             % constraint for SAM
Lf = 7;               % equalizer length - 1
L = Lc+Lf;            % number of terms in SAM - 1
P = L;                % number of terms in VCMA - 1
mu1 = 0.0002;         % SAM(AR) stepsize
mu2 = 0.0005;         % SAM(MA) stepsize
mu3 = 0.00002;        % VCMA stepsize
alpha = 1/20;         % forgetting factor for SAM(AR)
N = max(L,10);        % number of terms to average for SAM(MA)
f1 = zeros(Lf+1,1);
f1(ceil(Lf/2)) = 1;  % initialize SAM equalizer
%f1(1) = 1;
f2 = f1;
f3 = f1;

% transmitted data
s = sx*randn(iter,1);

% received data
x = filter(c.',1,s) + sn*randn(iter,1);

% allocate memory
y1 = zeros(iter,1);
y2 = zeros(iter,1);
y3 = zeros(iter,1);
A = zeros(L+Lf,1);  
B = zeros(L,1);
C = zeros(Lf+1,L);
Jsam1 = zeros(iter,1);
Jsam2 = zeros(iter,1);
Jsam3 = zeros(iter,1);
SSNR1 = zeros(iter,1);
SSNR2 = zeros(iter,1);
SSNR3 = zeros(iter,1);
MSE1 = zeros(iter,1);
MSE2 = zeros(iter,1);
MSE3 = zeros(iter,1);

% run both algorithms
hh = waitbar(0,'Equalizing the Gaussian source...');
%for n=P+Lf+1:iter
for n=N+L+Lf:iter
   % equalized data
   y1(n) = x(n:-1:n-Lf).'*f1;
   y3(n) = x(n:-1:n-Lf).'*f3;

   % SAM update (AR)

   % update A, B, C
   F = convmtx(f1',L);
   A = (1-alpha)*A + alpha*y1(n)*x(n-1:-1:n-L-Lf);
   B = F*A;
   C = (1-alpha)*C + alpha*x(n:-1:n-Lf) * y1(n-1:-1:n-L).';

   % form summation over l
   acc = zeros(Lf+1,1);  % accumulator
   for l=1:L
      acc = acc + B(l)*(A(l:l+Lf) + C(:,l));
   end

   % update f and renormalize
   f1 = f1 - mu1 * acc;
   if cnst == 0
      f1 = f1/norm(f1);  % enforce norm(f1) = 1
   elseif cnst == 1
      h1 = conv(c,f1);
      f1 = f1/norm(h1);  % enforce norm(h1) = 1
   else
      normh1 = sqrt(mean( y1(n:n-L-Lf).^2 ));
      f1 = f1/normh1;    % enforce norm(h1) = 1
   end

   % SAM update (MA)
   if mod(n,N) == 0
      % compute y(n) and y(n-l) values for this block
      for k=n-N+1-L:n
         y2(k) = x(k:-1:k-Lf).' * f2;
      end
      % form the outer summation
      acc0 = zeros(Lf+1,1);  % (accumulator)
      for l=1:L
         % form the two inner summations   
         acc1 = 0;  % (accumulators)
         acc2 = zeros(Lf+1,1);
         for k=n-N+1:n
            acc1 = acc1 + y2(k)*y2(k-l);
            xk   = x(k:-1:k-Lf);
            xkl  = x(k-l:-1:k-l-Lf);
            acc2 = acc2 + y2(k)*xkl + y2(k-l)*xk;
         end
         acc0 = acc0 + acc1 * acc2;
      end
      f2 = f2 - (mu2/N^2) * acc0;
      if cnst == 0
         f2 = f2/norm(f2);  % enforce norm(f2) = 1
      elseif cnst == 1
         h2 = conv(c,f2);
         f2 = f2/norm(h2);  % enforce norm(h2) = 1
      else
         normh2 = sqrt(mean( y2(n:n-L-Lf).^2 ));
         f2 = f2/normh2;    % enforce norm(h2) = 1
      end
   end  % SAM(MA) update

   % VCMA update
   X = fliplr(toeplitz( x(n-P:-1:n-Lf-P), x(n-P:n) ));
   Y = y3(n:-1:n-P);
   f3 = f3 - mu3*(sum(Y.^2)-rho)*(X*Y);

   % ---------- performance measurements ----------

   % SAM cost
   h1 = conv(c.',f1);
   Rhh = conv(h1,flipud(h1));
   Rhh = Rhh((length(Rhh)+1)/2 : end);  % make it one-sided
   Jsam1(n-L-Lf) = sum( (Rhh(2 : end)).^2 );

   h2 = conv(c.',f2);  
   Rhh = conv(h2,flipud(h2));
   Rhh = Rhh((length(Rhh)+1)/2 : end);  % make it one-sided
   Jsam2(n-L-Lf) = sum( (Rhh(2 : end)).^2 );

   h3 = conv(c.',f3);
   Rhh = conv(h3,flipud(h3));
   Rhh = Rhh((length(Rhh)+1)/2 : end);  % make it one-sided
   Jsam3(n-L-Lf) = sum( (Rhh(2 : end)).^2 );

   % SSNR
   [hpeak1,ind] = max(abs(h1));
   hout = h1; hout(ind) = 0;
   Eout1 = sum(hout.^2);
   SSNR1(n-P-Lf) = hpeak1^2/Eout1;

   [hpeak2,ind] = max(abs(h2));
   hout = h2; hout(ind) = 0;
   Eout2 = sum(hout.^2);
   SSNR2(n-P-Lf) = hpeak2^2/Eout2;

   [hpeak3,ind] = max(abs(h3));
   hout = h3; hout(ind) = 0;
   Eout3 = sum(hout.^2);
   SSNR3(n-P-Lf) = hpeak3^2/Eout3;

   % MSE
   MSE1(n-P-Lf) = (ssx*Eout1 + ssn*f1'*f1) / hpeak1^2;
   MSE2(n-P-Lf) = (ssx*Eout2 + ssn*f2'*f2) / hpeak2^2;
   MSE3(n-P-Lf) = (ssx*Eout3 + ssn*f3'*f3) / hpeak3^2;

   waitbar(n/iter,hh);
end  % time loop
close(hh);

% plot the results

time = 0:iter-Lf-P-1;
figure(1);
clf;
set(0,'DefaultAxesFontSize',16);
plot(time,Jsam1(time+1),'b-', time,Jsam2(time+1),'r-',...
   time,Jsam3(time+1),'k-','LineWidth',2);
title('SAM cost vs. symbol number','FontSize',16);
xlabel('symbol number','FontSize',16);
ylabel('J_S_A_M','FontSize',16);
legend('SAM(AR)','SAM(MA)','VCMA',1);
%axis([0 iter-Lf-P-1 0 1]);

figure(2);
clf;
set(0,'DefaultAxesFontSize',16);
plot(time,MSE1(time+1),'b-', time,MSE2(time+1),'r-',...
   time,MSE3(time+1),'k-','LineWidth',2);
title('MSE vs. symbol number','FontSize',16);
xlabel('symbol number','FontSize',16);
ylabel('MSE','FontSize',16);
legend('SAM(AR)','SAM(MA)','VCMA',1);
%axis([0 iter-Lf-P-1 0 1]);

figure(3);
clf;
%stem([0:length(c)-1],abs(c)/max(abs(c)),'g');
stem([0:length(c)-1],abs(c),'g*');
hold on
%stem([0:length(h1)-1]+1/4,abs(h1)/max(abs(h1)),'b');
%stem([0:length(h2)-1]+2/4,abs(h2)/max(abs(h2)),'r');
%stem([0:length(h3)-1]+3/4,abs(h3)/max(abs(h3)),'k');
stem([0:length(h1)-1]+1/4,abs(h1),'bo');
stem([0:length(h2)-1]+2/4,abs(h2),'rs');
stem([0:length(h3)-1]+3/4,abs(h3),'kd');
hold off
title('combined channels');

figure(4);
clf;
stem([0:length(f1)-1]+1/4,f1,'bo');
hold on
stem([0:length(f2)-1]+2/4,f2,'rs');
stem([0:length(f3)-1]+3/4,f3,'kd');
hold off
title('equalizers');
