function [f,Jsam,Jssnr,BR] = samiter(finit,r,c,params)
% [f,Jsam,Jssnr,BR] = samiter(finit,r,c,params)
%     Jsam = SAM cost vs. iteration number
%     Jssnr = 1/SSNR vs. iteration number
%     BR = achievable bit rate
%     f = final channel shortening filter (column vector)
%     finit = initialization of channel shortening filter
%     r = received data sequence (row vector)
%     c = channel (set to 1 if unknown...)
%
%     params = [offset,mu,N,nu,Lc,cnst,sx,sn,flag1]
%     offset = first data point of r to use (at least Lf+1+L)
%     mu = algorithm step size
%     N = data averaging size
%     nu = cyclic prefix length (window length - 1)
%     Lc = number of taps in channel - 1
%     cnst = contraint:
%        0 -> norm(f)=1
%        1 -> norm(h)=1, norm(h) assumed to be known exactly
%        2 -> norm(h)=1, norm(h) blindly estimated from the data
%     sx = sqrt(signal power)
%     sn = sqrt(noise power)
%     flag1: set this to zero to skip performance evaluation
%        0 -> skip non-essential calculations
%        1 -> do the calculations (for Jsam, Jssnr, BR)
%
%     SAM update routine
%     Written by Rick Martin, 05/17/2002
%     Last updated 06/20/2002
%     Cornell University Blind Equalization Research Group
%
%     Copyright 2002 Cornell University, All Rights Reserved
%
%     SOFTWARE shall mean "samiter.m", Cornell Research Foundation 
%     ("CRF") Docket D-3006.  SOFTWARE is made available to allow
%     non-commercial and research use of SOFTWARE.  Cornell
%     University specifically reserves all commercial rights to
%     SOFTWARE and these rights may be licensed by CRF on behalf   
%     of Cornell University to third parties.
%
%     CRF provides SOFTWARE on an ``as is'' basis.  CRF does not
%     warrant, guarantee, or make any representations regarding 
%     the use or results of SOFTWARE with respect to its
%     correctness, accuracy, reliability or performance.  The
%     entire risk of the use and performance of SOFTWARE is  
%     assumed by EVALUATOR.  ALL WARRANTIES INCLUDING, WITHOUT
%     LIMITATION, ANY WARRANTY OF FITNESS  FOR A PARTICULAR   
%     PURPOSE OR MERCHANTABILITY ARE HEREBY EXCLUDED.

% unpack the scalar parameters
offset = params(1);
mu     = params(2);
N      = params(3);
nu     = params(4);
Lc     = params(5);
cnst   = params(6);
sx     = params(7);
sn     = params(8);
flag1  = params(9);

ssx = sx*sx;
ssn = sn*sn;
Lf = length(finit)-1;
L = Lc + Lf;   % combined channel has L+1 taps
datasize = length(r) - (offset-1);
kmax = floor(datasize/N);
mu2 = 2*mu/(N*N);

y = zeros(size(r));
for n=(Lf+1):(offset-1)
   y(n) = r(n:-1:n-Lf) * finit;
end

f = finit/norm(finit);
Jsam = zeros(1,kmax);
Jssnr = zeros(1,kmax);
BR = zeros(1,kmax);
hh = waitbar(0,'Run, SAM, run...');
for k=0:kmax-1
   % compute y(n) values for this block
   for n=(k*N+offset):((k+1)*N-1+offset)
      y(n) = r(n:-1:n-Lf) * f;
   end

   % compute y(n-l) values for this block, nu+1<=l<=L
   for n=(k*N+offset-L):((k+1)*N-1+offset-(nu+1))
      y(n) = r(n:-1:n-Lf) * f;
   end

   % form the outer summation
   acc0 = zeros(Lf+1,1);  % (accumulator)
   for l=nu+1:L
      % form the two inner summations
      acc1 = 0;  % (accumulators)
      acc2 = zeros(Lf+1,1);
      for n=(k*N+offset):((k+1)*N-1+offset)
         acc1 = acc1 + y(n)*y(n-l);
         rn   = r(n:-1:n-Lf).';
         rnl  = r(n-l:-1:n-l-Lf).';
         acc2 = acc2 + y(n)*rnl + y(n-l)*rn;
      end
      acc0 = acc0 + acc1 * acc2;
   end

   % update f and renormalize
   f = f - mu2 * acc0;
   if cnst == 0
      f = f/norm(f);  % enforce norm(f) = 1
   elseif cnst == 1
      h = conv(c,f);
      f = f/norm(h);  % enforce norm(h) = 1
   else
      normh = sqrt(mean( y(n:n-L-Lf).^2 ));
      f = f/normh;    % enforce norm(h) = 1
   end

   % "non-essential" calculations
   if (flag1 == 1) | (n == nmax-1)
      h = conv(c,f);  % compute the combined channel

      % SAM cost for this equalizer
      Rhh = conv(h,fliplr(h));
      Rhh = Rhh((length(Rhh)+1)/2 : end);  % make it one-sided
      Jsam(k+1) = sum( (Rhh(nu+1 +1 : end)).^2 );

      % inverse shortening SNR for this equalizer
      Jssnr(k+1) = 1/ssnr2(h,nu);

      % find window in h with highest energy
      Ph = h.*h;
      sumh = zeros(1,length(Ph)-nu);
      for k2=1:length(Ph)-nu;
         sumh(k2) = sum(Ph(k2:k2+nu));
      end
      [dummy dopt] = max(sumh);
      dopt = dopt-1;  % optimal delay

      % achievable bit rate
      BR(k+1) = sum(bitrate(h,f,nu,dopt,ssx,ssn));
   end  % cost calculations

   waitbar((k+1)/kmax,hh);
end  % k loop
close(hh);
